@@ -1,4 +1,4 @@
-#![doc(html_root_url = "https://docs.rs/exmex/0.8.3")]
+#![doc(html_root_url = "https://docs.rs/exmex/0.8.4")]
 //! Exmex is a fast, simple, and **ex**tendable **m**athematical **ex**pression evaluator.  
 //! ```rust
 //! # use std::error::Error;
@@ -17,7 +17,7 @@
 //! ## Variables
 //! For variables we can use strings that are not in the list of operators as shown in the following expression.
 //! Additionally, variables should consist only of letters, numbers, and underscores. More precisely, they need to fit the
-//! regular expression 
+//! regular expression
 //! ```r"^[a-zA-Z_]+[a-zA-Z_0-9]*"```.
 //! Variables' values are passed as slices to [`eval`](FlatEx::eval).
 //! ```rust
@@ -97,8 +97,8 @@
 //! [`FromStr`](std::str::FromStr). In case the representation of your data type in the
 //! string does not match the number regex `r"\.?[0-9]+(\.[0-9]+)?"`, you have to pass a
 //! suitable regex and use the function
-//! [`parse_with_number_pattern`](parse::parse_with_number_pattern) instead of
-//! [`parse`](parse::parse). Here is an example for `bool`.
+//! [`parse_with_number_pattern`](parse_with_number_pattern) instead of
+//! [`parse`](parse). Here is an example for `bool`.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
@@ -132,7 +132,7 @@
 //! ## Priorities and Parentheses
 //! In Exmex-land, unary operators always have higher priority than binary operators, e.g.,
 //! `-2^2=4` instead of `-2^2=-4`. Moreover, we are not too strict regarding parentheses.
-//! For instance 
+//! For instance
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
@@ -147,8 +147,8 @@
 //!
 //! ## Display
 //!
-//! An instance of [`FlatEx`](FlatEx) can be displayed as string. Note that this 
-//! [`unparse`](FlatEx::unparse)d string does not necessarily coincide with the original 
+//! An instance of [`FlatEx`](FlatEx) can be displayed as string. Note that this
+//! [`unparse`](FlatEx::unparse)d string does not necessarily coincide with the original
 //! string. Especially, [`FlatEx`](FlatEx) does not keep track of variable names.
 //!
 //! ```rust
@@ -168,14 +168,19 @@
 //! future 😀.
 //!
 
+mod definitions;
 mod expression;
 mod operators;
-mod parse;
+mod parser;
 mod util;
 
-pub use expression::FlatEx;
+use std::{fmt::Debug, str::FromStr};
 
-pub use parse::{parse, parse_with_default_ops, parse_with_number_pattern, ExParseError};
+pub use expression::flat::FlatEx;
+use expression::{deep::DeepEx, flat};
+
+use num::Float;
+pub use parser::ExParseError;
 
 pub use operators::{make_default_operators, BinOp, Operator};
 
@@ -191,6 +196,79 @@ pub fn eval_str(text: &str) -> Result<f64, ExParseError> {
     Ok(flatex.eval(&[])?)
 }
 
+/// Parses a string and a vector of operators into an expression that can be evaluated.
+///
+/// # Errors
+///
+/// An error is returned in case [`parse_with_number_pattern`](parse_with_number_pattern)
+/// returns one.
+pub fn parse<'a, T>(text: &'a str, ops: &[Operator<'a, T>]) -> Result<FlatEx<'a, T>, ExParseError>
+where
+    <T as std::str::FromStr>::Err: Debug,
+    T: Copy + FromStr + Debug,
+{
+    let deepex = DeepEx::from_ops(text, ops)?;
+    Ok(flat::flatten(deepex))
+}
+
+/// Parses a string and a vector of operators and a regex pattern that defines the looks
+/// of a number into an expression that can be evaluated.
+///
+/// # Errors
+///
+/// An [`ExParseError`](ExParseError) is returned, if
+///
+//
+// from apply_regexes
+//
+/// * the argument `number_regex_pattern` cannot be compiled,
+/// * the argument `text` contained a character that did not match any regex (e.g.,
+///   if there is a `Δ` in `text` but no [operator](Operator) with
+///   [`repr`](Operator::repr) equal to `Δ` is given),
+//
+// from check_preconditions
+//
+/// * the to-be-parsed string is empty,
+/// * a number or variable is next to another one, e.g., `2 {x}`,
+/// * wlog a number or variable is on the right of a closing parenthesis, e.g., `)5`,
+/// * a binary operator is next to another binary operator, e.g., `2*/4`,
+/// * wlog a closing parenthesis is next to an opening one, e.g., `)(` or `()`,
+/// * too many closing parentheses at some position, e.g., `(4+6) - 5)*2`,
+/// * the last element is an operator, e.g., `1+`,
+/// * the number of opening and closing parenthesis do not match, e.g., `((4-2)`,
+//
+// from make_expression
+//
+/// * in `parsed_tokens` a closing parentheses is directly following an operator, e.g., `+)`, or
+/// * a unary operator is followed directly by a binary operator, e.g., `sin*`.
+///
+pub fn parse_with_number_pattern<'a, T>(
+    text: &'a str,
+    ops: &[Operator<'a, T>],
+    number_regex_pattern: &str,
+) -> Result<FlatEx<'a, T>, ExParseError>
+where
+    <T as std::str::FromStr>::Err: Debug,
+    T: Copy + FromStr + Debug,
+{
+    let deepex = DeepEx::from_pattern(text, ops, number_regex_pattern)?;
+    Ok(flat::flatten(deepex))
+}
+
+/// Parses a string into an expression that can be evaluated using default operators.
+///
+/// # Errors
+///
+/// An error is returned in case [`parse`](parse)
+/// returns one.
+pub fn parse_with_default_ops<'a, T>(text: &'a str) -> Result<FlatEx<'a, T>, ExParseError>
+where
+    <T as std::str::FromStr>::Err: Debug,
+    T: Float + FromStr + Debug,
+{
+    Ok(flat::flatten(DeepEx::from_str(text)?))
+}
+
 #[cfg(test)]
 mod tests {
 
@@ -199,8 +277,7 @@ mod tests {
     use crate::{
         eval_str,
         operators::{make_default_operators, BinOp, Operator},
-        parse::parse,
-        parse_with_default_ops,
+        parse, parse_with_default_ops,
         util::{assert_float_eq_f32, assert_float_eq_f64},
         ExParseError,
     };
@@ -216,7 +293,7 @@ mod tests {
             Ok(value)
         }
         fn readme_int() -> Result<u32, ExParseError> {
-            let ops = [
+            let ops = vec![
                 Operator {
                     repr: "|",
                     bin_op: Some(BinOp {
@@ -241,101 +318,96 @@ mod tests {
     }
     #[test]
     fn test_variables_curly() {
-        let operators = make_default_operators::<f64>();
-
         let to_be_parsed =
             "5*{x} + 4*log2(log(1.5-{gamma}))*({x}*-(tan(cos(sin(652.2-{gamma}))))) + 3*{x}";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 11.429314405093656);
         let to_be_parsed = "2*(4*{x} + y^2)";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.0]).unwrap(), 34.0);
 
         let to_be_parsed = "sin({myvwmlf4i58eo;w/-😕+sin(a)r_25})";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
         let to_be_parsed = "((sin({myvar_25})))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
     }
     #[test]
     fn test_variables() {
-        let operators = make_default_operators::<f64>();
-
         let to_be_parsed = "sin(sin(x - 1 / sin(y * 5)) + (5.0 - 1/z))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
-        let reference = |x: f64, y: f64, z: f64| {
-             ((x - 1.0 / (y * 5.0).sin()).sin() + (5.0 - 1.0 / z)).sin()
-        };
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
+        let reference =
+            |x: f64, y: f64, z: f64| ((x - 1.0 / (y * 5.0).sin()).sin() + (5.0 - 1.0 / z)).sin();
         assert_float_eq_f64(
             expr.eval(&[1.0, 2.0, 4.0]).unwrap(),
             reference(1.0, 2.0, 4.0),
         );
 
         let to_be_parsed = "0.02*sin(-(3*(2*(5.0 - 1/z))))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         let reference = |z: f64| 0.02 * (-(3.0 * (2.0 * (5.0 - 1.0 / z)))).sin();
         assert_float_eq_f64(expr.eval(&[4.0]).unwrap(), reference(4.0));
 
         let to_be_parsed = "y + 1 + 0.5 * x";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 3.0]).unwrap(), 3.5);
 
         let to_be_parsed = " -(-(1+x))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 2.0);
 
         let to_be_parsed = " sin(cos(-3.14159265358979*x))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), -0.841470984807896);
 
         let to_be_parsed = "5*sin(x * (4-y^(2-x) * 3 * cos(x-2*(y-1/(y-2*1/cos(sin(x*y))))))*x)";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5, 0.2532]).unwrap(), -3.1164569260604176);
 
         let to_be_parsed = "5*x + 4*y + 3*x";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 8.0);
 
         let to_be_parsed = "5*x + 4*y";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[0.0, 1.0]).unwrap(), 4.0);
 
         let to_be_parsed = "5*x + 4*y + x^2";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[2.5, 3.7]).unwrap(), 33.55);
         assert_float_eq_f64(expr.eval(&[12.0, 9.3]).unwrap(), 241.2);
 
         let to_be_parsed = "2*(4*x + y^2)";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.0]).unwrap(), 34.0);
 
         let to_be_parsed = "sin(myvar_25)";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
         let to_be_parsed = "((sin(myvar_25)))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
         let to_be_parsed = "(0 * myvar_25 + cos(x))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(
             expr.eval(&[1.5707963267948966, 3.141592653589793]).unwrap(),
             -1.0,
         );
 
         let to_be_parsed = "(-x^2)";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 1.0);
 
         let to_be_parsed = "log(x) + 2* (-x^2 + sin(4*y))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[2.5, 3.7]).unwrap(), 14.992794866624788);
 
         let to_be_parsed = "-sqrt(x)/(tanh(5-x)*2) + floor(2.4)* 1/asin(-x^2 + sin(4*sinh(y)))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(
             expr.eval(&[2.5, 3.7]).unwrap(),
             -2.5f64.sqrt() / (2.5f64.tanh() * 2.0)
@@ -343,15 +415,15 @@ mod tests {
         );
 
         let to_be_parsed = "asin(sin(x)) + acos(cos(x)) + atan(tan(x))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[0.5]).unwrap(), 1.5);
 
         let to_be_parsed = "sqrt(alpha^ceil(centauri))";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.1]).unwrap(), 4.0);
 
         let to_be_parsed = "trunc(x) + fract(x)";
-        let expr = parse(to_be_parsed, &operators).unwrap();
+        let expr = parse_with_default_ops::<f64>(to_be_parsed).unwrap();
         assert_float_eq_f64(expr.eval(&[23422.52345]).unwrap(), 23422.52345);
     }
 
@@ -375,7 +447,7 @@ mod tests {
 
     #[test]
     fn test_custom_ops() {
-        let custom_ops = [
+        let custom_ops = vec![
             Operator {
                 repr: "**",
                 bin_op: Some(BinOp {
