@@ -1,5 +1,5 @@
 use num::Float;
-use smallvec::{SmallVec, smallvec};
+use smallvec::{smallvec, SmallVec};
 
 /// Operators can be custom-defined by the library-user in terms of this struct.
 ///
@@ -35,7 +35,6 @@ pub struct Operator<'a, T> {
     pub unary_op: Option<fn(T) -> T>,
 }
 
-
 pub type VecOfUnaryFuncs<T> = SmallVec<[fn(T) -> T; 8]>;
 
 /// Container of unary operators of one expression
